#include <string.h>
#include <sodium.h>
#include <assert.h>

#include "../inc/lc6_crypto.h"
#include "../inc/lc6_helpers.h"

int lc6crypto_seeded = 0;

int lc6crypto_init(void) {
	return sodium_init();
}


LC6_USER* lc6crypto_genuserkey(LC6_USER *user) {
	assert(user);
	crypto_box_keypair(user->pub_key, user->priv_key);
	return user;
}

int lc6crypto_encrypt(
	const unsigned char *data,
	const int data_len,
	const unsigned char *skey,
	const unsigned char *dkey,
	unsigned char **enc_data,
	unsigned char **nonce)
{
	assert(enc_data);
	assert(nonce);

	*nonce = malloc(crypto_box_NONCEBYTES);
	assert(*nonce);
	memset(*nonce, 0, crypto_box_NONCEBYTES);

	*enc_data = malloc(crypto_box_MACBYTES + data_len);
	assert(*enc_data);
	memset(*enc_data, 0, crypto_box_MACBYTES + data_len);

	randombytes_buf(*nonce, crypto_box_NONCEBYTES);

	if ( crypto_box_easy(*enc_data, data, data_len, *nonce, dkey, skey) == -1 ) {
		free(*enc_data);
		free(*nonce);
		*enc_data = NULL;
		*nonce = NULL;
		return -1;
}

	return crypto_box_MACBYTES + data_len;
}

int lc6crypto_decrypt(
	const unsigned char *enc_data,
	const int enc_len,
	const unsigned char *skey,
	const unsigned char *dkey,
	unsigned char **data,
	const unsigned char *nonce)
{
	assert(enc_data);

	if ( enc_len < crypto_box_MACBYTES )
		return -1;

	*data = malloc(enc_len - crypto_box_MACBYTES);
	assert(*data);
	memset(*data, 0, enc_len - crypto_box_MACBYTES);

	if ( crypto_box_open_easy(*data, enc_data, enc_len, nonce, dkey, skey) != 0 ) {
		free(*data);
		*data = NULL;
		return -1;
	}

	return enc_len - crypto_box_MACBYTES;
}

int lc6crypto_readfile(
	const char *file,
	const unsigned char *password,
	unsigned char **data,
	int *data_len)
{
	FILE *fh;
	unsigned char key[crypto_secretbox_KEYBYTES];
	unsigned char *encrypted = NULL;
	size_t total = 0;

	crypto_generichash(key, sizeof(key), password, strlen((char*)password), NULL, 0);

	if ( ( fh = fopen(file, "r") ) == NULL )
		return -1;

	while(1) {
		unsigned char buf[4096];
		size_t len = 0;
		len = fread(buf, 1, sizeof(buf), fh);

		if ( len == 0 && feof(fh) ) {
			fclose(fh);
			if ( encrypted )
				free(encrypted);
			return -1;
		}

		if ( !encrypted )
			encrypted = malloc(len);
		else
			encrypted = realloc(encrypted, total + len);

		assert(encrypted);
		memcpy(encrypted + total, buf, len);

		total+=len;
	}

	fclose(fh);

	*data_len = total - crypto_secretbox_NONCEBYTES;
	*data = malloc(*data_len);
	assert(*data);
	memset(*data, '\0', total - crypto_secretbox_NONCEBYTES);

	if (crypto_secretbox_open_easy(*data, encrypted + crypto_secretbox_NONCEBYTES, *data_len, encrypted, key) != 0) {
		free(encrypted);
		free(*data);
		return -1;
	}


	free(encrypted);
	return 0;
}

int lc6crypto_writefile(
	const char *file,
	const unsigned char *password,
	const unsigned char *data,
	const int data_len)
{
	FILE *fh = NULL;
	unsigned char key[crypto_secretbox_KEYBYTES] = {0};
	unsigned char nonce[crypto_secretbox_NONCEBYTES] = {0};
	unsigned char *encrypted = NULL;
	int total_len = data_len + crypto_secretbox_NONCEBYTES + crypto_secretbox_MACBYTES;

	crypto_generichash(key, sizeof(key), password, strlen((char*)password), NULL, 0);
	randombytes_buf(nonce, sizeof(nonce));

	encrypted = malloc(total_len);
	assert(encrypted);
	memset(encrypted, '\0', total_len);
	memcpy(encrypted, nonce, crypto_secretbox_NONCEBYTES);

	crypto_secretbox_easy(encrypted + crypto_secretbox_NONCEBYTES, data, data_len, nonce, key);

	if ( ( fh = fopen(file, "w") ) == NULL )
		return -1;

	if ( fwrite(encrypted, total_len, 1, fh) != 1 ) {
		free(encrypted);
		fclose(fh);
		return -1;
	}

	fclose(fh);
	free(encrypted);
	return 0;
}

unsigned char* lc6crypto_hash(
	unsigned char *data,
	int            data_len)
{
	unsigned char *hash = malloc(crypto_generichash_BYTES);

	assert(hash);

	crypto_generichash(hash, sizeof hash, data, data_len, NULL, 0);
	return hash;
	
}

void lc6crypto_random(unsigned char *data, int data_len) {
	assert(data);
	randombytes_buf(data, data_len);
}
